from rest_framework.generics import get_object_or_404

from mayan.apps.acls.models import AccessControlList

from ..models.workflow_models import Workflow


class ParentObjectWorkflowTemplateAPIViewMixin:
    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['workflow_template'] = self.workflow_template
        return context

    def get_workflow_template(self, permission=None):
        queryset = self.get_workflow_template_queryset()

        if not permission:
            permission = getattr(
                self, 'mayan_external_object_permissions', {}
            ).get(self.request.method, (None,))[0]

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs['workflow_template_id']
        )

    def get_workflow_template_queryset(self):
        return Workflow.objects.all()

    def initial(self, *args, **kwargs):
        result = super().initial(*args, **kwargs)
        # Ensure self.workflow_template is initialized to allow the
        # browseable API view to display when attempting to introspect
        # the serializer and the parent object is not found.
        self.workflow_template = None

        self.workflow_template = self.get_workflow_template()
        return result


class ParentObjectWorkflowTemplateStateAPIViewMixin(
    ParentObjectWorkflowTemplateAPIViewMixin
):
    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['workflow_template_state'] = self.workflow_template_state
        return context

    def get_workflow_template_state(self, permission=None):
        queryset = self.get_workflow_template_state_queryset()

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs['workflow_template_state_id']
        )

    def get_workflow_template_state_queryset(self):
        return self.get_workflow_template().states.all()

    def initial(self, *args, **kwargs):
        result = super().initial(*args, **kwargs)
        # Ensure self.workflow_template_state is initialized to allow the
        # browseable API view to display when attempting to introspect
        # the serializer and the parent object is not found.
        self.workflow_template_state = None

        self.workflow_template_state = self.get_workflow_template_state()
        return result


class ParentObjectWorkflowTemplateTransitionAPIViewMixin(
    ParentObjectWorkflowTemplateAPIViewMixin
):
    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['workflow_template_transition'] = self.workflow_template_transition
        return context

    def get_workflow_template_transition(self, permission=None):
        queryset = self.get_workflow_template_transition_queryset()

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return get_object_or_404(
            queryset=queryset, pk=self.kwargs[
                'workflow_template_transition_id'
            ]
        )

    def get_workflow_template_transition_queryset(self):
        return self.get_workflow_template().transitions.all()

    def initial(self, *args, **kwargs):
        result = super().initial(*args, **kwargs)
        # Ensure self.workflow_template_transition is initialized to allow the
        # browseable API view to display when attempting to introspect
        # the serializer and the parent object is not found.
        self.workflow_template_transition = None

        self.workflow_template_transition = self.get_workflow_template_transition()
        return result



from io import StringIO
from unittest import mock

from django.core import management

from mayan.apps.cabinets.tests.mixins import CabinetTestMixin
from mayan.apps.document_indexing.tests.mixins import IndexTemplateTestMixin
from mayan.apps.storage.utils import NamedTemporaryFile
from mayan.apps.testing.tests.base import BaseTestCase


class MountCabinetManagementCommandTestCase(CabinetTestMixin, BaseTestCase):
    auto_create_test_cabinet = True

    def _call_command(self, mount_point, label):
        stderr = StringIO()
        stdout = StringIO()
        management.call_command(
            'mirroring_mount_cabinet', label, mount_point, stderr=stderr,
            stdout=stdout
        )
        return stderr.getvalue(), stdout.getvalue()

    @mock.patch('fuse.FUSE', autospec=True)
    def test_cabinet_mount_basic(self, FUSE):
        with NamedTemporaryFile() as file_object:
            FUSE.return_value = None
            self._call_command(
                label=self._test_cabinet.label, mount_point=file_object.name
            )

    @mock.patch('fuse.FUSE', autospec=True)
    def test_cabinet_mount_with_invalid_slug(self, FUSE):
        with NamedTemporaryFile() as file_object:
            FUSE.return_value = None
            with self.assertRaises(expected_exception=SystemExit):
                self._call_command(
                    label='invalid', mount_point=file_object.name
                )


class MountIndexManagementCommandTestCase(
    IndexTemplateTestMixin, BaseTestCase
):
    auto_create_test_index_template = False

    def _call_command(self, mount_point, slug):
        stderr = StringIO()
        stdout = StringIO()
        management.call_command(
            'mirroring_mount_index', slug, mount_point, stderr=stderr,
            stdout=stdout
        )
        return stderr.getvalue(), stdout.getvalue()

    def setUp(self):
        super().setUp()
        self._create_test_index_template()

    @mock.patch('fuse.FUSE', autospec=True)
    def test_index_mount_basic(self, FUSE):
        with NamedTemporaryFile() as file_object:
            FUSE.return_value = None
            self._call_command(
                mount_point=file_object.name,
                slug=self._test_index_template.slug
            )

    @mock.patch('fuse.FUSE', autospec=True)
    def test_index_mount_with_invalid_slug(self, FUSE):
        with NamedTemporaryFile() as file_object:
            FUSE.return_value = None
            with self.assertRaises(expected_exception=SystemExit):
                self._call_command(
                    mount_point=file_object.name, slug='invalid'
                )



from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0023_auto_20150715_0259'),
    ]

    operations = [
        migrations.AlterField(
            model_name='documentversion',
            name='comment',
            field=models.TextField(
                default='', verbose_name='Comment', blank=True
            ),
            preserve_default=True,
        ),
    ]



from rest_framework.filters import BaseFilterBackend, OrderingFilter

from mayan.apps.acls.models import AccessControlList


class MayanObjectPermissionsFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        permission = self.get_mayan_object_permissions(request=request, view=view)

        if permission:
            return AccessControlList.objects.restrict_queryset(
                queryset=queryset, permission=permission,
                user=request.user
            )
        else:
            return queryset

    def get_mayan_object_permissions(self, request, view):
        try:
            return getattr(view, 'get_mayan_object_permissions')()
        except AttributeError:
            return getattr(
                view, 'mayan_object_permissions', {}
            ).get(request.method, (None,))[0]


class MayanSortingFilter(OrderingFilter):
    ordering_param = '_ordering'

    def get_default_valid_fields(self, queryset, view, context={}):
        return ()



from unittest import mock

from mayan.apps.documents.tests.base import GenericDocumentTestCase

from .mixins import DocumentVersionOCRTaskTestMixin


class OCRTaskTestCase(
    DocumentVersionOCRTaskTestMixin, GenericDocumentTestCase
):
    @mock.patch(
        'mayan.apps.ocr.events.event_ocr_document_version_finished.commit'
    )
    def test_task_document_version_ocr_finished_error_create(
        self, event_ocr_document_version_finished_commit
    ):
        TEST_EXCEPTION_TEXT = 'Fake exception'

        self._silence_logger(name='mayan.apps.ocr.tasks')

        event_ocr_document_version_finished_commit.side_effect = Exception(
            TEST_EXCEPTION_TEXT
        )

        self._clear_events()

        with self.assertRaises(expected_exception=Exception):
            self._execute_task_document_version_ocr_finished()

        self.assertEqual(
            self._test_document_version.error_log.count(), 1
        )
        self.assertEqual(
            self._test_document_version.error_log.first().text,
            TEST_EXCEPTION_TEXT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import hashlib
import io
import logging

import cairosvg
from furl import furl
from PIL import Image

from django.conf import settings
from django.db import models
from django.shortcuts import reverse
from django.utils.encoding import force_bytes
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.validators import validate_internal_name
from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_models import Document
from mayan.apps.events.classes import (
    EventManagerMethodAfter, EventManagerSave
)
from mayan.apps.events.decorators import method_event
from mayan.apps.file_caching.models import Cache, CachePartitionFile
from mayan.apps.templating.classes import Template

from .events import (
    event_signature_capture_created, event_signature_capture_deleted,
    event_signature_capture_edited,
)
from .literals import STORAGE_NAME_SIGNATURE_CAPTURES_CACHE
from .managers import ValidSignatureCaptureManager

logger = logging.getLogger(name=__name__)


class SignatureCapture(ExtraDataModelMixin, models.Model):
    document = models.ForeignKey(
        on_delete=models.CASCADE, related_name='signature_captures',
        to=Document, verbose_name=_('Document')
    )
    data = models.TextField(
        blank=True, help_text=_(
            'Data representing the handwritten signature.'
        ), verbose_name=_('Signature capture data')
    )
    svg = models.TextField(
        blank=True, help_text=_(
            'Vector representation of the handwritten signature.'
        ), verbose_name=_('SVG signature capture data')
    )
    text = models.CharField(
        help_text=_('Print version of the captured signature.'),
        max_length=224, verbose_name=_('Text')
    )
    user = models.ForeignKey(
        on_delete=models.CASCADE, related_name='signature_captures',
        to=settings.AUTH_USER_MODEL, verbose_name=_('User')
    )
    date_time_created = models.DateTimeField(
        auto_now_add=True, db_index=True,
        verbose_name=_('Date and time created')
    )
    date_time_edited = models.DateTimeField(
        auto_now=True, db_index=True,
        verbose_name=_('Date and time edited')
    )
    internal_name = models.CharField(
        db_index=True, help_text=_(
            'This value will be used when referencing this signature '
            'capture in relationship to the document. Can only contain '
            'letters, numbers, and underscores.'
        ), max_length=255, validators=[validate_internal_name],
        verbose_name=_('Internal name')
    )

    objects = models.Manager()
    valid = ValidSignatureCaptureManager()

    class Meta:
        ordering = ('-date_time_created',)
        unique_together = ('document', 'internal_name')
        verbose_name = _('Signature capture')
        verbose_name_plural = _('Signature captures')

    def __str__(self):
        return '{} - {}'.format(self.text, self.get_date_time_created())

    @cached_property
    def cache(self):
        return Cache.objects.get(
            defined_storage_name=STORAGE_NAME_SIGNATURE_CAPTURES_CACHE
        )

    @cached_property
    def cache_partition(self):
        partition, created = self.cache.partitions.get_or_create(
            name='{}'.format(self.pk)
        )
        return partition

    @method_event(
        action_object='self',
        event=event_signature_capture_deleted,
        event_manager_class=EventManagerMethodAfter,
        target='document'
    )
    def delete(self, *args, **kwargs):
        self.cache_partition.delete()
        return super().delete(*args, **kwargs)

    def generate_image(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        # The parameters 'maximum_layer_order',
        # `transformation_instance_list`, `user` are not used, but added
        # to retain interface compatibility.
        cache_filename = '{}'.format(self.get_hash())

        try:
            self.cache_partition.get_file(filename=cache_filename)
        except CachePartitionFile.DoesNotExist:
            logger.debug(
                'signature capture cache file "%s" not found', cache_filename
            )

            image = self.get_image()
            with io.BytesIO() as image_buffer:
                image.save(image_buffer, format='PNG')

                with self.cache_partition.create_file(filename=cache_filename) as file_object:
                    file_object.write(image_buffer.getvalue())
        else:
            logger.debug(
                'signature_capture cache file "%s" found', cache_filename
            )

        return cache_filename

    def get_absolute_url(self):
        return reverse(
            viewname='signature_captures:signature_capture_detail', kwargs={
                'signature_capture_id': self.pk
            }
        )

    def get_api_image_url(self, *args, **kwargs):
        final_url = furl()
        final_url.args = kwargs
        final_url.path = reverse(
            viewname='rest_api:signature_capture-image',
            kwargs={
                'document_id': self.document.pk,
                'signature_capture_id': self.pk
            }
        )
        final_url.args['_hash'] = self.get_hash()

        return final_url.tostr()

    def get_date_time_created(self):
        return Template(
            template_string='{{ instance.date_time_created }}'
        ).render(
            context={'instance': self}
        )
    get_date_time_created.short_description = _('Creation date and time')

    def get_hash(self):
        return hashlib.sha256(force_bytes(self.svg)).hexdigest()

    def get_image(self):
        stream = io.BytesIO()
        cairosvg.svg2png(url=self.svg, write_to=stream)
        image = Image.open(stream)

        return image

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'actor': 'user',
            'action_object': 'document',
            'event': event_signature_capture_created,
            'target': 'self'
        },
        edited={
            'action_object': 'document',
            'event': event_signature_capture_edited,
            'target': 'self'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)
